//// Fast Network Simplex for Approximate W_2 distance -- test unit
//// by Nicolas Papadakis (Nov. 2018) from code of Nicolas Bonneel (Nov. 2013)



#include <iostream>
#include <vector>

#include "network_simplex_simple.h"
#include <chrono>

#include <stdio.h>
#define DEBUG 0

using namespace lemon;

// all types should be signed
typedef int64_t arc_id_type; // {short, int, int64_t} ; Should be able to handle (n1*n2+n1+n2) with n1 and n2 the number of nodes (INT_MAX = 46340^2, I64_MAX = 3037000500^2)
typedef double supply_type; // {float, double, int, int64_t} ; Should be able to handle the sum of supplies and *should be signed* (a demand is a negative supply)
typedef double cost_type;  // {float, double, int, int64_t} ; Should be able to handle (number of arcs * maximum cost) and *should be signed*

struct TsFlow {
    int from, to;
    supply_type amount;
};

template<typename T>
cost_type sqrdistancend(T* a, T* b,int dim) {
    switch(dim){
        case 2:
            return (a[0] - b[0])*(a[0] - b[0]) + (a[1] - b[1])*(a[1] - b[1]);
        case 3:
            return (a[0] - b[0])*(a[0] - b[0]) + (a[1] - b[1])*(a[1] - b[1])+ (a[2] - b[2])*(a[2] - b[2]);
        default:
            return (a[0] - b[0])*(a[0] - b[0]);
    }
}

double compute_EMD(int n1, double *pos_1, supply_type *weights1,int n2, double *pos_2, supply_type *weights2, int dim){
    typedef FullBipartiteDigraph Digraph;
    DIGRAPH_TYPEDEFS(FullBipartiteDigraph);
    
    Digraph di(n1, n2);
    
    NetworkSimplexSimple<Digraph, supply_type, cost_type, arc_id_type> net(di, true, (int)(n1 + n2), (int64_t)(n1*n2));
    
    net.supplyMap(weights1, n1, weights2, n2);
    
    
    arc_id_type idarc = 0;
    for (int i = 0; i < n1; i++) {
        for (int j = 0; j < n2; j++) {
            cost_type d = sqrdistancend(&pos_1[i * dim], &pos_2[j * dim],dim);
            net.setCost(di.arcFromId(idarc), d);
            idarc++;
        }
    }
    
    // We run the network simplex
    
    int ret = net.run();
    double resultdist = net.totalCost();
    return resultdist ;
}


double compute_approximate_EMD(int n1, double *pos_1, supply_type *weights1,int n2, double *pos_2, supply_type *weights2, int dim,int p0, int max_iter, double stop_crit){
    typedef FullTransShipmentgraph TSgraph;
    TSGRAPH_TYPEDEFS(FullTransShipmentgraph);
    
    bool random_initialization=0;
    
    int p=p0;
    int pp=(int)pow((double)p0+0.1,1./dim);
    if(pp<2)
        pp=2;
    
    if (random_initialization==0)
        p=pow(pp,dim);
    
    
    std::vector<double> flow_ts((n1+n2)*p),coordsZ(dim*p);
    srand(time(NULL));
    
    
    if (random_initialization==1) {
        // Random initialization
        for (int i = 0; i < (int)p/2; i++){
            int index=int((rand()*double(n1-1)) / (double)RAND_MAX);
            
            for(int kk=0;kk<dim;kk++)
                coordsZ[i*dim+kk] = pos_1[index*dim+kk];
        }
        
        for (int i = (int)p/2; i < p; i++){
            int index=int((rand()*double(n2-1)) / (double)RAND_MAX);
            
            for(int kk=0;kk<dim;kk++)
                coordsZ[i*dim+kk] = pos_2[index*dim+kk];
        }
    }
    else{
        //to initialize regularly the positions z, we consider that there are p=pp^dim intermediate locations with pp>1
        std::vector<double> max_pos(dim);
        std::vector<double> min_pos(dim);
        
        for (int i = 0; i < dim; i++){
            max_pos[i]=-1e15;
            min_pos[i]=1e15;
        }
        for (int i = 0; i < n1; i++)
            for(int k=0;k<dim;k++){
                if(max_pos[k]<pos_1[i*dim+k])
                    max_pos[k]=pos_1[i*dim+k];
                if(min_pos[k]>pos_1[i*dim+k])
                    min_pos[k]=pos_1[i*dim+k];
            }
        
        for (int i = 0; i < n2; i++)
            for(int k=0;k<dim;k++){
                if(max_pos[k]<pos_2[i*dim+k])
                    max_pos[k]=pos_2[i*dim+k];
                if(min_pos[k]>pos_2[i*dim+k])
                    min_pos[k]=pos_2[i*dim+k];
            }
        // printf("Min/Max values for each dimension:\n");
        //  for(int k=0;k<dim;k++)
        //      printf("%f %f\n",min_pos[k],max_pos[k]);
        
        
        for(int i=0;i<p;i++){
            int ii=i;
            for(int k=0;k<dim;k++){
                int kk=ii/(pow(pp,dim-1-k));
                double tt=((double)kk+1.)/((double)pp+1.);
                coordsZ[i*dim+k] =min_pos[k]+tt*(max_pos[k]-min_pos[k]);
                ii=ii-kk*pow(pp,dim-1-k);
            }
        }
        /*printf("Initial Z positions:\n");
         for(int i=0;i<p;i++){
         for(int k=0;k<dim;k++)
         printf("%f ", coordsZ[i*dim+k]);
         printf("\n");
         }*/
        
        
        
    }
    
    
    
    
    
    
    
    
    
    double resultdist_ts =100000000.;
    int new_p=p;
    
    
    
    
    for(int iter=0;iter<max_iter;iter++) {
        if (new_p<p)
            p=new_p;
        
        TSgraph ts(n1,p, n2);
        
        
        NetworkSimplexSimple<TSgraph, supply_type, cost_type, arc_id_type> net_ts(ts, false, n1 + n2+p, (n1+n2)*p);
        
        
        
        
        net_ts.supplyMap(weights1, n1, p, weights2, n2);
        
        arc_id_type idarc = 0;
        for (int i = 0; i < n1; i++) {
            for (int k = 0; k < p; k++) {
                cost_type d = sqrdistancend(&pos_1[i * dim], &coordsZ[k * dim],dim);
                net_ts.setCost(ts.arcFromId(idarc), d);
                idarc++;
            }
        }
        for (int k = 0; k < p; k++) {
            for (int j = 0; j < n2; j++) {
                cost_type d = sqrdistancend(&coordsZ[k * dim], &pos_2[j * dim],dim);
                net_ts.setCost(ts.arcFromId(idarc), d);
                idarc++;
            }
        }
        
        
        // We run the network simplex
        
        int ret_ts = net_ts.run();
        double old_resultdist_ts=resultdist_ts;
        resultdist_ts = net_ts.totalCost();  // resultdist is the EMD
        if(DEBUG==1)
            printf("Intern cost %f\n",resultdist_ts);
        if ((old_resultdist_ts-resultdist_ts)/old_resultdist_ts<stop_crit){
            if(DEBUG==1)
                printf("Stopped at iteration %d\n",iter);
            break;
        }
        
        new_p=0;
        //Update the transshipment locations
        for (arc_id_type k = 0; k < p; k++) {
            cost_type tmp=0.;
            for (arc_id_type i = 0; i < n1; i++) {
                flow_ts[i*p + k]=net_ts.flow(ts.arcFromId(i*p + k));
                tmp+= flow_ts[i*p + k];
            }
            for (arc_id_type j = 0; j < n2; j++){
                flow_ts[n1*p+k*n2 + j]=net_ts.flow(ts.arcFromId(n1*p+k*n2 + j));
                tmp+= flow_ts[n1*p+k*n2 + j];
            }
            
            if (tmp > 1E-18){
                
                for(int kk=0;kk<dim;kk++){
                    coordsZ[new_p * dim+kk]=0.;
                    
                    for (arc_id_type i = 0; i < n1; i++){
                        coordsZ[new_p * dim+kk]+= pos_1[i* dim+kk]*flow_ts[i*p + k];
                    }
                    for (arc_id_type j = 0; j < n2; j++){
                        coordsZ[new_p * dim+kk]+= pos_2[j* dim+kk]*flow_ts[n1*p+k*n2 + j];
                    }
                    coordsZ[new_p * dim+kk]/=tmp;
                }
                new_p++;
            }
        }
        
        
    }
    
    

    
    //std::vector<double> resultdist_k(p);
    resultdist_ts=0.;
    std::vector<int> valid_i(n1), valid_j(n2);
    
    for (int k=0;k<p;k++){
        int n1_k=0;
        int n2_k=0;
        for (int i=0;i<n1;i++) {
            if (flow_ts[i*p + k]> 1E-18){
                valid_i[i]=1;
                n1_k++;
            }
            else
                valid_i[i]=0;
            
        }
        for (int j=0;j<n2;j++){
            if (flow_ts[n1*p+k*n2 + j]> 1E-18){
                valid_j[j]=1;
                n2_k++;
            }
            else
                valid_j[j]=0;
        }
        
        if(n1_k>0 && n2_k>0) {
            std::vector<supply_type> weights1_k(n1_k), weights2_k(n2_k);
            std::vector<double> pos1_k(n1_k*dim), pos2_k(n2_k*dim);
            
            int cur=0;
            for (int i=0;i<n1;i++){
                if (valid_i[i]==1) {
                    weights1_k[cur]=flow_ts[i*p + k];
                    for(int kk=0;kk<dim;kk++)
                        pos1_k[cur*dim+kk]=pos_1[i*dim + kk];
                    cur++;
                }
            }
            
            cur=0;
            for (int j=0;j<n2;j++){
                if (valid_j[j]==1) {
                    weights2_k[cur]=-flow_ts[n1*p+k*n2 + j];
                    for(int kk=0;kk<dim;kk++)
                        pos2_k[cur*dim+kk]=pos_2[j*dim + kk];
                    cur++;
                }
            }
            if (n1_k+n2_k<2000)
                resultdist_ts+=compute_EMD(n1_k, &pos1_k[0], &weights1_k[0],n2_k,&pos2_k[0],&weights2_k[0],dim);
            else
                resultdist_ts+=compute_approximate_EMD(n1_k, &pos1_k[0], &weights1_k[0],n2_k,&pos2_k[0],&weights2_k[0],dim,p0,max_iter,stop_crit);
            
        }
    }
    
    return resultdist_ts;
    
}


int main() {
    
    
    srand(time(NULL));
    
    arc_id_type n1 = pow(2,12), n2 = n1, p=16; // just demo for the case n1=n2 ; adapt otherwise
    printf("Problem of dimension %d\n",(int)n1);
    int dim=2; //space dimension
    
    
    std::vector<double> coordsXY(n1 * dim);
    std::vector<supply_type> weights1(n1), weights2(n2); // works faster with integer weights though
    
    
    
    for (int i = 0; i < n1 * dim; i++) {
        coordsXY[i] =double(rand()*100.) / (double)RAND_MAX;
        // coordsXY[i] = double(i)/double(n1-1);
    }
    
    
    // random supplies and demands
    for (int i = 0; i < n1; i++){
       // weights1[i]= (double)rand() / (double)RAND_MAX;
        weights1[i]= std::exp(-pow((coordsXY[i]-50),2.));
    }
    
    
    double s = 0.;
    for (int i = 0; i < n1; i++) {
        s += weights1[i];
    }
    
    
    for (int i = 0; i < n1; i++) {
        weights1[i]/=s;   // we normalize the input histogram
    }
    
    
    for (int i = 0; i < n2; i++) {
        weights2[i]= -(double)rand() / (double)RAND_MAX;
        // weights2[i]=-exp(-pow((coordsXY[i] - 0.25)/ 0.1,2.)) - exp(-pow((coordsXY[i] - 0.75)/ 0.1,2.));
        
    }
    
    
    s = 0.;
    for (int i = 0; i < n2; i++){
        s += fabs(weights2[i]);
    }
    
    
    for (int i = 0; i < n2; i++) {
        weights2[i]/= s;  // we normalize the target histogram
    }
    
    
    std::chrono::high_resolution_clock::time_point startTime = std::chrono::high_resolution_clock::now();
    
    //Compute EMD. This is not the last optimized version of the code of N. Bonneel that deals with sparse matrices. The sparse version is not used here since it is counterproductive for the transshipment problem. Running time should therefore be compared with the code in https://github.com/nbonneel/network_simplex for dimensions> 10^4.
    
    
    double resultdist =sqrt(compute_EMD(n1, &coordsXY[0], &weights1[0],n2,&coordsXY[0],&weights2[0],dim));
    
    std::chrono::high_resolution_clock::time_point endTime = std::chrono::high_resolution_clock::now();
    
    double time_OT=(double)std::chrono::duration_cast<std::chrono::microseconds>(endTime - startTime).count() / 1000000.;
    
    std::cout << "EMD = "<<resultdist  << " in " << time_OT << "s." <<  std::endl;
    
    
    std::chrono::high_resolution_clock::time_point startTime_ts = std::chrono::high_resolution_clock::now();
    
    
    int max_iter=100;
    double stop_crit=1e-2;
    //compute approximate EMD
    double resultdist_ts =sqrt(compute_approximate_EMD(n1, &coordsXY[0], &weights1[0],n2,&coordsXY[0],&weights2[0],dim,p,max_iter,stop_crit));
    
    
    
    std::chrono::high_resolution_clock::time_point endTime_ts = std::chrono::high_resolution_clock::now();
    
    double time_AOT=(double)std::chrono::duration_cast<std::chrono::microseconds>(endTime_ts - startTime_ts).count() / 1000000.;
    
    std::cout << "EMD Approx = "<<resultdist_ts << " in " << time_AOT << "s." << std::endl;
    
    
    
    
    return 0;
}



