/* This file is a matlab mex function for computing the  transportation barycenter
 * between two vectors given a cost matrix.
 *
 * It was written by Nicolas Papadakis (2018)
 * It is based on the code written by Antoine Rolet (2014) that mainly consists of a wrapper
 * of the code written by Nicolas Bonneel available on this page
 *          http://people.seas.harvard.edu/~nbonneel/FastTransport/
 *
 * Please give relevant credit to the original author (Nicolas Bonneel) if
 * you use this code for a publication.
 *
 */
#include "mex.h"
#include <iostream>
#include <vector>

#include "network_simplex_simple.h"


using namespace lemon;
typedef unsigned int node_id_type;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    int n, m, p, i,     j, n1, n2,cur;
    double *X, *Y, *DX,*DY, *demd, *optimalx,*optimaly, max,max_iter;
    const mwSize *dimx, *dimy, *dimdX, *dimdY, *dim_tmp;
    
    
    typedef FullTransShipmentgraph TSgraph;
    DIGRAPH_TYPEDEFS(FullTransShipmentgraph);
    
    // Check if the number of input and output argument is correc
    if (nrhs < 3)
        mexErrMsgTxt("There should be at least 3 input arguments");
    if (nlhs > 3)
        mexErrMsgTxt("Too many output argument");
    if (nlhs < 3)
        mexErrMsgTxt("Not enough output argument");
    
    // Get the input arguments X, Y, DX, (DY), (max_iter)
 
    dimx = mxGetDimensions(prhs[0]);
    dimy = mxGetDimensions(prhs[1]);
    dimdX = mxGetDimensions(prhs[2]);
    X = mxGetPr(prhs[0]);
    Y = mxGetPr(prhs[1]);
    DX = mxGetPr(prhs[2]);
    
   
    int given_DY=0;
    
    
    if (nrhs >3) {
        dim_tmp = mxGetDimensions(prhs[3]);
        if (dim_tmp[0]==1 && dim_tmp[1]==1)  //it should be the number of iterations
            max_iter=*mxGetPr(prhs[3]);
        else   //otheriwise the fourth input is DY
            given_DY=1;
    }
    
    
    if(given_DY==1) {  //read DY
        dimdY = mxGetDimensions(prhs[3]);
        DY = mxGetPr(prhs[3]);
    }
    else {     //Set DY=DX' if DY not provided
        if (dimx[0]!=dimy[0])
            mexErrMsgTxt("The input vectors should be of the same size if no ground matrix DY is provided");
        DY = mxGetPr(prhs[2]);
        dimdY = mxGetDimensions(prhs[2]);
        
        
        
    }
    
    if(nrhs>4)
        max_iter=*mxGetPr(prhs[4]);
    else
        max_iter=-1;
    // Check dimensions
    
    if (dimx[1] != 1 || dimy[1] != 1)
        mexErrMsgTxt("First two arguments should be column vectors");

    
 
    
    n1 = dimdX[0];
    if (dimx[0] != n1)
        mexErrMsgTxt(
                     "Ground metric matrix Dx first dimension should be the length of the first vector X");
    
    n2 = dimdY[1];
    if (given_DY==0)//DY'
        n2 = dimdY[0];
    
    if (dimy[0] != n2)
        mexErrMsgTxt(
                     "Ground metric matrix Dy second dimension should be the length of the second vector Y");
    
    p=dimdY[0];
    if (given_DY==0) //DY'
        p = dimdY[1];
    
    if (dimdX[1] != p)
        mexErrMsgTxt(
                     "Second dimension of Dx and first  dimension of Dy  should be the same");
    
    
    
    
    
    // Get the number of non zero coordinates for r and c
    n=0;
    for (node_id_type i=0; i<n1; i++) {
        double val=*(X+i);
        if (val>0) {
            n++;
        }
    }
    m=0;
    for (node_id_type i=0; i<n2; i++) {
        double val=*(Y+i);
        if (val>0) {
            m++;
        }
    }

    
    // Define the graph
    
    std::vector<int> indI(n), indJ(m), indK(p);
    std::vector<double> weightsX(n), weightsY(m), weightsZ(p);
    TSgraph di(n,p,m);
    NetworkSimplexSimple<TSgraph,double,double, node_id_type> net(di, true, n+m+p, (n+m)*p,max_iter);

    // Set supply and demand, don't account for 0 values (faster)
    
    max=0;
    cur=0;
    for (node_id_type i=0; i<n1; i++) {
        double val=*(X+i);
        if (val>0) {
            weightsX[ di.nodeFromId(cur) ] = val;
            max+=val;
            indI[cur++]=i;
        }
    }
    
    // Demand is actually negative supply...
    
    max=0;
    cur=0;
    for (node_id_type i=0; i<n2; i++) {
        double val=*(Y+i);
        if (val>0) {
            weightsY[ di.nodeFromId(cur) ] = -val;
            indJ[cur++]=i;
            
            max-=val;
        }
    }
    
    // Intermediate node (for barycenter)
    cur=0;
    for (node_id_type i=0; i<p; i++) {
            weightsZ[ di.nodeFromId(cur) ] = 0.;
            indK[cur++]=i;
    }

    net.supplyMap(&weightsX[0], n, &weightsZ[0], p, &weightsY[0], m);
   
    // Set the cost of each edge
    max=0;
    for (node_id_type i=0; i<n; i++) {
        for (node_id_type k=0; k<p; k++) {
            double val=*(DX+indI[i]+indK[k]*n1);
            net.setCost(di.arcFromId(i*p+k), val);
            if (val>max) {
                max=val;
            }
        }
    }
    
    int index_layer=n*p;
    max=0;
    for (node_id_type k=0; k<p; k++) {
        for (node_id_type j=0; j<m; j++) {
            double val=*(DY+indK[k]+indJ[j]*p);
            if (given_DY==0)   //DY'
                val=*(DY+indJ[j]+indK[k]*n2);
            net.setCost(di.arcFromId(index_layer+k*m+j),val);//+j*p+k), val);//
            if (val>max) {
                max=val;
            }
        }
    }

    
    // Solve the problem with the network simplex algorithm
    //mexPrintf("coucou");
    int ret=net.run();
    
    if (ret!=(int)net.OPTIMAL) {
        if (ret==(int)net.INFEASIBLE) {
            mexErrMsgTxt("Infeasible problem");
        }
        if (ret==(int)net.UNBOUNDED) {
            mexErrMsgTxt("Unbounded problem");
        }
    }
    
    // Define the output arguments
    
    plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
    demd = mxGetPr(plhs[0]);
    
    plhs[1] = mxCreateDoubleMatrix(n1,p, mxREAL);
    optimalx = mxGetPr(plhs[1]);
    plhs[2] = mxCreateDoubleMatrix(p,n2, mxREAL);
    optimaly = mxGetPr(plhs[2]);
  
    // Set the output values
    
     for (node_id_type i=0; i<n; i++) {
            for (node_id_type k=0; k<p; k++)
            {
                *(optimalx+indI[i]+indK[k]*n1) = net.flow(di.arcFromId(i*p+k));
            }
        }
    
    
        for (node_id_type k=0; k<p; k++) {
            for (node_id_type j=0; j<m; j++)
            {
                *(optimaly+indK[k]+indJ[j]*p) = net.flow(di.arcFromId(index_layer+k*m+j));//+j*p+k));//
            }
        }
    
    
    *demd=net.totalCost();
    
}
