function [cost, gamma]=compute_Approximate_WP(pos_x,mu_x,pos_y,mu_y,p,d_z_ref,compute_gamma)
%
% Multi-scale approximation of WP distance  (%Algorithm 3 in the paper)
%
% Inputs:
% pos_x         : dirac locations of the first density. Matrix (d_x , dim)
% mu_x          : dirac weights. Vector (d_x , 1)
% pos_y         : dirac locations of the second density. Matrix (d_y , dim)
% mu_y          : dirac weights. Vector (d_y , 1)
% d_z           : number of diracs to approximate the barycenter. Positive integer
% p             : p-Wasserstin distance. Scalar >1
% compute_gamma : will return an approximation of the transport matrix if set to 1. boolean

% Outputs:
% cost : the approximate p-Wasserstein cost between (x,mu_x) and (y,mu_y)
% gamma: the corresponding approximation of the optimal matrix (if
% compute_gamma=1)


max_iter=100;
stop_crit=1e-3;

cost=1e10;
gamma=[];
if min(mu_x(:))<0
    disp('error: mu_x should have non negative values')
    return
end

if min(mu_y(:))<0
    disp('error: mu_y should have non negative values')
    return
end

if abs(sum(mu_x(:))-sum(mu_y(:)))>1e-15
    disp('normalization: mu_x and mu_y should have the same mass')
    mu_y=mu_y*sum(mu_x(:))/sum(mu_y(:));
end

d_z=d_z_ref;

if d_z<1
    disp('error: d_z should be greater than 1')
    return
else
    d_z=uint16(d_z);
end

d_x=length(mu_x);


if size(pos_x,1)~=d_x
    disp('error: length of input histogram mu_x and first dimension of pos_x should be the same')
    return
end

d_y=length(mu_y);
if size(pos_y,1)~=d_y
    disp('error: length of input histogram mu_y and first dimension of pos_y should be the same')
    return
end

dim=size(pos_x,2);
if size(pos_y,2)~=dim
    disp('error: set of points in pos_x and pos_y should be of the same dimension')
    return
end




if p<1
    disp('error: p has to be greater than 1')
    return
end




random_initialization=0;

d_zz=max(2,uint16(double(d_z+0.1)^(1./dim)));
if random_initialization==0
   d_z=d_zz^dim;
end
   



pos_z=zeros(d_z,dim);


if random_initialization
    %Random Initialization of barycenter positions from positions of inputs
    index_z=randsample(1:d_x,d_z/2,true,mu_x);
    pos_z(1:d_z/2,:)=pos_x(index_z,:)+rand(size(d_z/2,2))*0.01;
    
    index_z=randsample(1:d_y,d_z-d_z/2,true,mu_y);
    pos_z(d_z/2+1:end,:)=pos_y(index_z,:)+rand(size(d_z-d_z/2,2))*0.01;
    
else

       max_pos=max(max(pos_x, [],1),max(pos_y, [],1));
       min_pos=max(min(pos_x, [],1),min(pos_y, [],1));     

 
        
        for k=1:d_z
            cpt=k;
            for kk=1:dim
                ll=floor(double(cpt-1)/double((d_zz)^(dim-kk)));
                tt=(double(ll+1))/(double(d_zz)+1.);
                pos_z(k,kk)=min_pos(kk)+tt*(max_pos(kk)-min_pos(kk));
                cpt=cpt-ll*(d_zz^(dim-kk));
            end
        end

end



gamma_x=[];
gamma_y=[];

Cx=[];
Cy=[];

d_z_new=d_z;
%Algorithm 2 in the paper
for i=1:max_iter
    d_z=d_z_new;  %in case locations have been removed

    C_xzd= @(pos_z,d)  repmat(pos_z(:,d)',[d_x 1])-repmat(pos_x(:,d),[1 d_z]);
    C_yzd= @(pos_z,d)  repmat(pos_z(:,d),[1 d_y])-repmat(pos_y(:,d)',[d_z 1]);
    cost_old=cost;
                                                         
    %update cost matrices
    Cx = abs(C_xzd(pos_z,1)).^p;
    Cy = abs(C_yzd(pos_z,1)).^p;
    for d=2:dim
        Cx = Cx+ abs(C_xzd(pos_z,d)).^p;
        Cy = Cy+ abs(C_yzd(pos_z,d)).^p;
    end
                                                         
    % Compute barycenter between mu_x and mu_y on the grid z by solving the transshipment problem
    [cost, gamma_x, gamma_y]=mexBAR(mu_x,mu_y,Cx,Cy);
    
    
    if abs(cost-cost_old)/cost_old<stop_crit
        break
    end
                                                         
    % Update the latent positions z
    normalize=repmat(sum(gamma_x,1)'+sum(gamma_y,2), [1 dim]);
    
    % Detect empty locations
    active=normalize>1e-15;
    
    normalize=max(1e-15,normalize);
    switch(p)        
        case 1 %explicit median  expression
            for d=1:dim
                pos_z(:,d)=wmedian(gamma_x,gamma_y,pos_x(:,d),pos_y(:,d));
            end
        case 2  %explicit mean expression
            pos_z=(gamma_x'*pos_x+gamma_y*pos_y)./normalize;
            
        otherwise   
            if p>2
                %do 10 Newton iterations
                
                for kk=1:10
                    for d=1:dim
                        fx=C_xzd(pos_z,d);  
                        fy=C_yzd(pos_z,d);
                        
                        ffx=gamma_x.*abs(fx).^(p-2);
                        ffy=gamma_y.*abs(fy).^(p-2);
                   
                        fp=sum(ffx.*fx,1)'+sum(ffy.*fy,2);
                        fpp=sum(ffx,1)'+sum(ffy,2);
                        
                        pos_z=pos_z-fp./((p-1)*(fpp));
                        %seems OK: Reduce to z=z-(gamma_x'*(z-x)+gamma_y*(z-y)) in case p=2
                        
                    end
                end
            else   %p<2
                %do 10 "median" linearization
                if p<1.5  %initialize with median
                    for d=1:dim
                        pos_z(:,d)=wmedian(gamma_x,gamma_y,pos_x(:,d),pos_y(:,d));
                    end
                else  %initialize with mean
                    pos_z=(gamma_x'*pos_x+gamma_y*pos_y)./normalize;
                end
                for kk=1:10
                    for d=1:dim
                        fx=abs(C_xzd(pos_z,d)).^(p-1);
                        fy=abs(C_yzd(pos_z,d)).^(p-1);
                        pos_z(:,d)=wmedian(gamma_x.*fx,gamma_y.*fy,pos_x(:,d),pos_y(:,d));
                        
                    end
                end
                
            end
    end
                           
    % Remove empty locations
    d_z_new=sum(active(:))/dim;
    pos_z0=zeros(d_z_new,dim);
    pos_z0(:)=pos_z(active);
    pos_z=pos_z0;
end


%Untie the strings.
w=gamma_x'*ones(d_x,1);
cost_k=zeros(d_z,1);
gamma=sparse(size(gamma_x,1),size(gamma_y,2));
gamma_k=[];
                           
for k=1:d_z   %to process in parallel if parpool available
    gamma_k{k}=[];
    index_xk{k}=[];
    index_yk{k}=[];
end

parfor k=1:d_z
    vect=w*0;
    vect(k)=1.;
    mu_xkt=gamma_x*vect;
    mu_ykt=gamma_y'*vect;
    index_xk{k}=mu_xkt>1e-15;
    index_yk{k}=mu_ykt>1e-15;
    
    pos_xk=pos_x;
    pos_yk=pos_y;
    dim_xk=sum(index_xk{k}(:));
    dim_yk=sum(index_yk{k}(:));
    
    mu_xk=mu_xkt(index_xk{k});
    mu_yk=mu_ykt(index_yk{k});
    
    %ensure compatible mass (may have issues when very low mass)
    masse_xk=sum(mu_xk(:));
    masse_yk=sum(mu_yk(:));    
    masse_k=(masse_xk+masse_yk)/2;
    mu_xk=mu_xk/masse_xk*masse_k;
    mu_yk=mu_yk/masse_yk*masse_k;
                                                                 
                                                                 
    if dim_xk+dim_yk<2000   %low dimensional problem: solve exactly (%Algorithm 1 in the paper)
        [costk, gammak]=compute_WP(pos_xk(index_xk{k},:),mu_xk,pos_yk(index_yk{k},:),mu_yk,p);
        cost_k(k)=costk;
        if compute_gamma
            gamma_k{k}=gammak(:);
        end
    else      %apply the multi-scale process recursively
        [costk, gammak]=compute_Approximate_WP(pos_xk(index_xk{k},:),mu_xk,pos_yk(index_yk{k},:),mu_yk,p,d_z_ref,compute_gamma);
        cost_k(k)=costk;
        if compute_gamma
            gamma_k{k}=gammak(:);
        end
        
    end
    
end
                           
                           
if compute_gamma   %build the transport matrix
    for k=1:d_z
        index_k{k}=sparse(index_xk{k}*index_yk{k}');
        gamma(index_k{k}>0)=gamma_k{k}(:);
    end
end

cost=sum(cost_k(:));



